@@ -117,17 +117,6 @@ let is_ground_env evd env =
   List.for_all is_ground_rel_decl (rel_context env) &&
   List.for_all is_ground_named_decl (named_context env)
 
-(* Memoization is safe since evar_map and environ are applicative
-   structures *)
-let memo f =
-  let module E = Ephemeron.K2 in
-  let m = E.create () in
-  fun x y -> match E.get_key1 m, E.get_key2 m with
-  | Some x', Some y' when x == x' && y == y' -> Option.get (E.get_data m)
-  | _ -> let r = f x y in E.set_key1 m x; E.set_key2 m y; E.set_data m r; r
-
-let is_ground_env = memo is_ground_env
-
 (* Return the head evar if any *)
 
 exception NoHeadEvar
@@ -308,15 +297,25 @@ let update_var src tgt subst =
       let csubst_var = Id.Map.add id (Constr.mkVar tgt) subst.csubst_var in
       { subst with csubst_var; csubst_rev }
 
+module VarSet =
+struct
+  type t = Id.t -> bool
+  let empty _ = false
+  let full _ = true
+  let variables env id = is_section_variable env id
+end
+
+let program_naming =
+  Goptions.declare_bool_option_and_ref ~depr:true
+    ~key:["Program";"Naming"] ~value:false
+
 type naming_mode =
-  | KeepUserNameAndRenameExistingButSectionNames
-  | KeepUserNameAndRenameExistingEvenSectionNames
-  | KeepExistingNames
+  | RenameExistingBut of VarSet.t
   | FailIfConflict
-  | ProgramNaming
+  | ProgramNaming of VarSet.t
 
 let push_rel_decl_to_named_context
-  ?(hypnaming=KeepUserNameAndRenameExistingButSectionNames)
+  ~hypnaming
   sigma decl ((subst, avoid, nc) : ext_named_context) =
   let open EConstr in
   let open Vars in
@@ -346,37 +345,42 @@ let push_rel_decl_to_named_context
         using this function. For now, we only attempt to preserve the
         old behaviour of Program, but ultimately, one should do something
         about this whole name generation problem. *)
-    if hypnaming = ProgramNaming then next_name_away na avoid
-    else
+    match hypnaming with
+    | ProgramNaming _ when program_naming () -> next_name_away na avoid
+    | ProgramNaming _ | RenameExistingBut _ | FailIfConflict ->
       (* id_of_name_using_hdchar only depends on the rel context which is empty
          here *)
       next_ident_away (id_of_name_using_hdchar empty_env sigma (RelDecl.get_type decl) na) avoid
   in
   match extract_if_neq id na with
-  | Some id0 when hypnaming = KeepUserNameAndRenameExistingEvenSectionNames ||
-                  (hypnaming = KeepUserNameAndRenameExistingButSectionNames ||
-                   hypnaming = ProgramNaming) &&
-                  not (is_section_variable (Global.env ()) id0) ->
-      (* spiwack: if [id<>id0], rather than introducing a new
-          binding named [id], we will keep [id0] (the name given
-          by the user) and rename [id0] into [id] in the named
-          context. Unless [id] is a section variable. *)
-      let subst = update_var id0 id subst in
-      let d = decl |> NamedDecl.of_rel_decl (fun _ -> id0) |> map_decl (csubst_subst subst) in
-      let nc = replace_var_named_declaration id0 id nc in
-      let avoid = Id.Set.add id (Id.Set.add id0 avoid) in
-      (push_var id0 subst, avoid, push_named_context_val d nc)
-  | Some id0 when hypnaming = FailIfConflict ->
+  | Some id0 ->
+    begin match hypnaming with
+    | RenameExistingBut f | ProgramNaming f ->
+      if f id0 then
+        (* spiwack: if [id0] is a section variable renaming it is
+            incorrect. We revert to a less robust behaviour where
+            the new binder has name [id]. Which amounts to the same
+            behaviour than when [id=id0]. *)
+        let d = decl |> NamedDecl.of_rel_decl (fun _ -> id) |> map_decl (csubst_subst subst) in
+        (push_var id subst, Id.Set.add id avoid, push_named_context_val d nc)
+      else
+        (* spiwack: if [id<>id0], rather than introducing a new
+            binding named [id], we will keep [id0] (the name given
+            by the user) and rename [id0] into [id] in the named
+            context. Unless [id] is a section variable. *)
+        let subst = update_var id0 id subst in
+        let d = decl |> NamedDecl.of_rel_decl (fun _ -> id0) |> map_decl (csubst_subst subst) in
+        let nc = replace_var_named_declaration id0 id nc in
+        let avoid = Id.Set.add id (Id.Set.add id0 avoid) in
+        (push_var id0 subst, avoid, push_named_context_val d nc)
+    | FailIfConflict ->
        user_err Pp.(Id.print id0 ++ str " is already used.")
-  | _ ->
-      (* spiwack: if [id0] is a section variable renaming it is
-          incorrect. We revert to a less robust behaviour where
-          the new binder has name [id]. Which amounts to the same
-          behaviour than when [id=id0]. *)
-      let d = decl |> NamedDecl.of_rel_decl (fun _ -> id) |> map_decl (csubst_subst subst) in
-      (push_var id subst, Id.Set.add id avoid, push_named_context_val d nc)
+    end
+  | None ->
+    let d = decl |> NamedDecl.of_rel_decl (fun _ -> id) |> map_decl (csubst_subst subst) in
+    (push_var id subst, Id.Set.add id avoid, push_named_context_val d nc)
 
-let push_rel_context_to_named_context ?hypnaming env sigma typ =
+let push_rel_context_to_named_context ~hypnaming env sigma typ =
   (* compute the instances relative to the named context and rel_context *)
   let open EConstr in
   let inst_vars = EConstr.identity_subst_val (named_context_val env) in
@@ -389,7 +393,7 @@ let push_rel_context_to_named_context ?hypnaming env sigma typ =
     (* with vars of the rel context *)
     (* We do keep the instances corresponding to local definition (see above) *)
     let (subst, _, env) =
-      Context.Rel.fold_outside (fun d acc -> push_rel_decl_to_named_context ?hypnaming sigma d acc)
+      Context.Rel.fold_outside (fun d acc -> push_rel_decl_to_named_context ~hypnaming sigma d acc)
         (rel_context env) ~init:(empty_csubst, avoid, named_context_val env) in
     (env, csubst_subst subst typ, inst_rels@inst_vars, subst)
 
@@ -437,7 +441,11 @@ let new_pure_evar ?(src=default_source) ?(filter = Filter.identity) ?identity
 (* Converting the env into the sign of the evar to define *)
 let new_evar ?src ?filter ?abstract_arguments ?candidates ?naming ?typeclass_candidate
     ?principal ?hypnaming env evd typ =
-  let sign,typ',instance,subst = push_rel_context_to_named_context ?hypnaming env evd typ in
+  let hypnaming = match hypnaming with
+  | Some n -> n
+  | None -> RenameExistingBut (VarSet.variables (Global.env ()))
+  in
+  let sign,typ',instance,subst = push_rel_context_to_named_context ~hypnaming env evd typ in
   let map c = csubst_subst subst c in
   let candidates = Option.map (fun l -> List.map map l) candidates in
   let instance =
@@ -512,7 +520,7 @@ let restrict_evar evd evk filter ?src candidates =
   | Some [] -> raise (ClearDependencyError (*FIXME*)(Id.of_string "blah", (NoCandidatesLeft evk), None))
   | _ -> Evd.restrict evk filter ?candidates ?src evd
 
-let rec check_and_clear_in_constr env evdref err ids global c =
+let rec check_and_clear_in_constr ~is_section_variable env evdref err ids ~global c =
   (* returns a new constr where all the evars have been 'cleaned'
      (ie the hypotheses ids have been removed from the contexts of
      evars). [global] should be true iff there is some variable of [ids] which
@@ -536,7 +544,7 @@ let rec check_and_clear_in_constr env evdref err ids global c =
             (* If evk is already defined we replace it by its definition *)
             let nc = Evd.existential_value !evdref (EConstr.of_existential ev) in
             let nc = EConstr.Unsafe.to_constr nc in
-              (check_and_clear_in_constr env evdref err ids global nc)
+            check_and_clear_in_constr ~is_section_variable env evdref err ids ~global nc
           else
             (* We check for dependencies to elements of ids in the
                evar_info corresponding to e and in the instance of
@@ -567,12 +575,12 @@ let rec check_and_clear_in_constr env evdref err ids global c =
                 ctxt l (Id.Map.empty,[]) in
             (* Check if some rid to clear in the context of ev has dependencies
                in the type of ev and adjust the source of the dependency *)
-            let _nconcl =
+            let _nconcl : Constr.t =
               try
                 let nids = Id.Map.domain rids in
-                let global = Id.Set.exists (fun id -> is_section_variable (Global.env ()) id) nids in
+                let global = Id.Set.exists is_section_variable nids in
                 let concl = EConstr.Unsafe.to_constr (evar_concl evi) in
-                check_and_clear_in_constr env evdref (EvarTypingBreak ev) nids global concl
+                check_and_clear_in_constr ~is_section_variable env evdref (EvarTypingBreak ev) nids ~global concl
               with ClearDependencyError (rid,err,where) ->
                 raise (ClearDependencyError (Id.Map.find rid rids,err,where)) in
 
@@ -587,7 +595,7 @@ let rec check_and_clear_in_constr env evdref err ids global c =
               evdref := evd;
               Evd.existential_value0 !evdref ev
 
-      | _ -> Constr.map (check_and_clear_in_constr env evdref err ids global) c
+      | _ -> Constr.map (check_and_clear_in_constr ~is_section_variable env evdref err ids ~global) c
 
 let clear_hyps_in_evi_main env sigma hyps terms ids =
   (* clear_hyps_in_evi erases hypotheses ids in hyps, checking if some
@@ -595,13 +603,14 @@ let clear_hyps_in_evi_main env sigma hyps terms ids =
      the contexts of the evars occurring in evi *)
   let evdref = ref sigma in
   let terms = List.map EConstr.Unsafe.to_constr terms in
-  let global = Id.Set.exists (fun id -> is_section_variable (Global.env ()) id) ids in
+  let is_section_variable id = is_section_variable (Global.env ()) id in
+  let global = Id.Set.exists is_section_variable ids in
   let terms =
-    List.map (check_and_clear_in_constr env evdref (OccurHypInSimpleClause None) ids global) terms in
+    List.map (check_and_clear_in_constr ~is_section_variable env evdref (OccurHypInSimpleClause None) ids ~global) terms in
   let nhyps =
     let check_context decl =
       let err = OccurHypInSimpleClause (Some (NamedDecl.get_id decl)) in
-      NamedDecl.map_constr (check_and_clear_in_constr env evdref err ids global) decl
+      NamedDecl.map_constr (check_and_clear_in_constr ~is_section_variable env evdref err ids ~global) decl
     in
     let check_value vk = match force_lazy_val vk with
     | None -> vk
@@ -618,6 +627,15 @@ let clear_hyps_in_evi_main env sigma hyps terms ids =
   in
   (!evdref, nhyps,List.map EConstr.of_constr terms)
 
+let check_and_clear_in_constr env evd err ids c =
+  let evdref = ref evd in
+  let c = EConstr.Unsafe.to_constr c in
+  let _ : constr = check_and_clear_in_constr
+      ~is_section_variable:(fun _ -> true) ~global:true
+      env evdref err ids c
+  in
+  !evdref
+
 let clear_hyps_in_evi env sigma hyps concl ids =
   match clear_hyps_in_evi_main env sigma hyps [concl] ids with
   | (sigma,nhyps,[nconcl]) -> (sigma,nhyps,nconcl)
@@ -628,6 +646,14 @@ let clear_hyps2_in_evi env sigma hyps t concl ids =
   | (sigma,nhyps,[t;nconcl]) -> (sigma,nhyps,t,nconcl)
   | _ -> assert false
 
+let evar_nodes_of_term c =
+  let rec evrec acc c =
+    match kind c with
+    | Evar (n, l) -> Evar.Set.add n (List.fold_left evrec acc l)
+    | _ -> Constr.fold evrec acc c
+  in
+  evrec Evar.Set.empty (EConstr.Unsafe.to_constr c)
+
 (* spiwack: a few functions to gather evars on which goals depend. *)
 let queue_set q is_dependent set =
   Evar.Set.iter (fun a -> Queue.push (is_dependent,a) q) set
@@ -671,7 +697,7 @@ let gather_dependent_evars q evm =
 
 let gather_dependent_evars evm l =
   let q = Queue.create () in
-  List.iter (fun a -> Queue.add (false,a) q) l;
+  List.iter (queue_term q false) l;
   gather_dependent_evars q evm
 
 (* /spiwack *)
@@ -806,8 +832,8 @@ let occur_evar_upto sigma n c =
 
 let judge_of_new_Type evd =
   let open EConstr in
-  let (evd', s) = new_univ_variable univ_rigid evd in
-  (evd', { uj_val = mkType s; uj_type = mkType (Univ.super s) })
+  let (evd', s) = new_sort_variable univ_rigid evd in
+  (evd', { uj_val = mkSort s; uj_type = mkSort (Sorts.super s) })
 
 let subterm_source evk ?where (loc,k) =
   let evk = match k with
@@ -833,7 +859,7 @@ let compare_cumulative_instances cv_pb variances u u' sigma =
   match Evd.add_constraints sigma cstrs with
   | sigma ->
     Inl (Evd.add_universe_constraints sigma soft)
-  | exception Univ.UniverseInconsistency p -> Inr p
+  | exception UGraph.UniverseInconsistency p -> Inr p
 
 let compare_constructor_instances evd u u' =
   let open UnivProblem in
@@ -862,9 +888,8 @@ let eq_constr_univs_test ~evd ~extended_evd t u =
   let eq_sorts s1 s2 =
     if Sorts.equal s1 s2 then true
     else
-      let u1 = Sorts.univ_of_sort s1 and u2 = Sorts.univ_of_sort s2 in
-      try sigma := add_universe_constraints !sigma UnivProblem.(Set.singleton (UEq (u1, u2))); true
-      with Univ.UniverseInconsistency _ | UniversesDiffer -> false
+      try sigma := add_universe_constraints !sigma UnivProblem.(Set.singleton (UEq (s1, s2))); true
+      with UGraph.UniverseInconsistency _ | UniversesDiffer -> false
   in
   let kind1 = kind_of_term_upto evd in
   let kind2 = kind_of_term_upto extended_evd in
