@@ -103,6 +103,7 @@ let make_coqtop_args fname =
         with
         | None -> "", base_args
         | Some proj ->
+          Minilib.log ("About to read project file " ^ proj);
           let warning_fn x = Feedback.msg_warning Pp.(str x) in
           proj, coqtop_args_from_project (read_project_file ~warning_fn proj) @ base_args
   in
@@ -1178,13 +1179,6 @@ let apply_unicode_binding =
   cb_on_current_term (fun t ->
     t.script#apply_unicode_binding())
 
-let mem_usage =
-  cb_on_current_term (fun t ->
-    let open Gc in
-    let stats = Gc.stat () in
-    Printf.printf "live_words = %d heap_words = %d top_heap_words = %d\n%!"
-      stats.live_words stats.heap_words stats.top_heap_words)
-
 let coqtop_arguments sn =
   init_bpts sn;
   let dialog = GWindow.dialog ~title:"Coqtop arguments" () in
@@ -1368,6 +1362,16 @@ let emit_to_focus window sgn =
 (** {2 Creation of the main coqide window } *)
 
 let build_ui () =
+  (* issue 12779 *)
+  GtkData.StyleContext.add_provider_for_screen
+    (Gdk.Screen.default ())
+    begin
+      let provider = GObj.css_provider () in
+      provider#load_from_data "button { box-shadow: none }";
+      provider#as_css_provider
+    end
+    GtkData.StyleContext.ProviderPriority.application;
+
   let w = GWindow.window
     ~wmclass:("CoqIDE","CoqIDE")
     ~width:window_width#get ~height:window_height#get
@@ -1618,8 +1622,6 @@ let build_ui () =
       ~callback:(fun _ -> on_current_term (fun sn ->
          sn.messages#default_route#clear;
          sn.messages#default_route#add_string (MicroPG.get_documentation ())));
-    item "Memory usage" ~label:"Memory usage"  (* temporary, for debugging *)
-      ~callback:MiscMenu.mem_usage;
     item "About Coq" ~label:"_About" ~stock:`ABOUT
       ~callback:MiscMenu.about
   ];
@@ -1778,6 +1780,20 @@ let main files =
     in the path. Note that the -coqtop option to coqide overrides
     this default coqtop path *)
 
+let coqide_specific_usage = Boot.Usage.{
+  executable_name = "coqide";
+  extra_args = "";
+  extra_options = "\n\
+CoqIDE specific options:\
+\n  -f _CoqProjectFile         set _CoqProject file to _CoqProjectFile\
+\n  -unicode-bindings f1 .. f2 load files f1..f2 with extra unicode bindings\
+\n  -coqtop dir                look for coqidetop in dir\
+\n  -coqtop-flags              extra flags for the coqtop subprocess\
+\n  -debug                     enable debug mode\
+\n  -xml-debug                 enable debug mode and print XML messages to/from CoqIDE\
+\n"
+}
+
 let read_coqide_args argv =
   let set_debug () =
     CDebug.set_debug_all true;
@@ -1791,17 +1807,17 @@ let read_coqide_args argv =
     |"-coqtop" :: prog :: args ->
       if coqtop = None then filter_coqtop (Some prog) project_files bindings_files out args
       else (output_string stderr "Error: multiple -coqtop options"; exit 1)
+    |"-coqtop" :: [] ->
+      output_string stderr "Error: missing argument after -coqtop"; exit 1
     |"-f" :: file :: args ->
       if project_files <> None then
         (output_string stderr "Error: multiple -f options"; exit 1);
       let d = CUnix.canonical_path_name (Filename.dirname file) in
       let warning_fn x = Format.eprintf "%s@\n%!" x in
       let p = CoqProject_file.read_project_file ~warning_fn file in
-      filter_coqtop coqtop (Some (d,p)) out bindings_files args
+      filter_coqtop coqtop (Some (d,p)) bindings_files out args
     |"-f" :: [] ->
       output_string stderr "Error: missing project file name"; exit 1
-    |"-coqtop" :: [] ->
-      output_string stderr "Error: missing argument after -coqtop"; exit 1
     |"-debug"::args ->
       set_debug ();
       filter_coqtop coqtop project_files bindings_files out args
@@ -1813,12 +1829,17 @@ let read_coqide_args argv =
       Coq.ideslave_coqtop_flags := Some flags;
       filter_coqtop coqtop project_files bindings_files out args
     | ("-v" | "--version") :: _ ->
+      (* This does the same thing as Usage.version () but printed differently *)
       Printf.printf "CoqIDE, version %s\n" Coq_config.version;
+      (* Unlike coqtop we don't accumulate queries so we exit immediately *)
       exit 0
-    |arg::args when out = [] && CString.is_prefix "-psn_" arg ->
+    | ("-h"|"-H"|"-?"|"-help"|"--help") :: _ ->
+      Boot.Usage.print_usage stderr coqide_specific_usage;
+      exit 0
+    | arg::args when out = [] && CString.is_prefix "-psn_" arg ->
       (* argument added by MacOS during .app launch *)
       filter_coqtop coqtop project_files bindings_files out args
-    |arg::args -> filter_coqtop coqtop project_files bindings_files (arg::out) args
+    | arg::args -> filter_coqtop coqtop project_files bindings_files (arg::out) args
     |[] -> (coqtop,project_files,bindings_files,List.rev out)
   in
   let coqtop,project_files,bindings_files,argv = filter_coqtop None None [] [] argv in
