@@ -76,24 +76,28 @@ let ide_doc = ref None
 let get_doc () = Option.get !ide_doc
 let set_doc doc = ide_doc := Some doc
 
-let add ((s,eid),(sid,verbose)) =
+let add ((((s,eid),(sid,verbose)),off),(line_nb,bol_pos)) =
   let doc = get_doc () in
-  let pa = Pcoq.Parsable.make (Stream.of_string s) in
+  let open Loc in
+  (* note: this won't yield correct values for bol_pos_last,
+     but the debugger doesn't use that *)
+  let loc = { (initial ToplevelInput) with bp=off; line_nb } in
+  let pa = Pcoq.Parsable.make ~loc (Stream.of_string s) in
   match Stm.parse_sentence ~doc sid ~entry:Pvernac.main_entry pa with
   | None -> assert false (* s may not be empty *)
   | Some ast ->
     let doc, newid, rc = Stm.add ~doc ~ontop:sid verbose ast in
     set_doc doc;
-    let rc = match rc with `NewTip -> CSig.Inl () | `Unfocus id -> CSig.Inr id in
+    let rc = match rc with Stm.NewAddTip -> CSig.Inl () | Stm.Unfocus id -> CSig.Inr id in
     ide_cmd_warns ~id:newid ast;
     (* All output is sent through the feedback mechanism rather than stdout/stderr *)
-    newid, (rc, "")
+    newid, rc
 
 let edit_at id =
   let doc = get_doc () in
   match Stm.edit_at ~doc id with
-  | doc, `NewTip -> set_doc doc; CSig.Inl ()
-  | doc, `Focus { Stm.start; stop; tip} -> set_doc doc; CSig.Inr (start, (stop, tip))
+  | doc, Stm.NewTip -> set_doc doc; CSig.Inl ()
+  | doc, Stm.Focus { Stm.start; stop; tip} -> set_doc doc; CSig.Inr (start, (stop, tip))
 
 (* TODO: the "" parameter is a leftover of the times the protocol
  * used to include stderr/stdout output.
@@ -113,10 +117,10 @@ let annotate phrase =
   let doc = get_doc () in
   let pa = Pcoq.Parsable.make (Stream.of_string phrase) in
   match Stm.parse_sentence ~doc (Stm.get_current_state ~doc) ~entry:Pvernac.main_entry pa with
-  | None -> Richpp.richpp_of_pp 78 (Pp.mt ())
+  | None -> Richpp.richpp_of_pp ~width:78 (Pp.mt ())
   | Some ast ->
     (* XXX: Width should be a parameter of annotate... *)
-    Richpp.richpp_of_pp 78 (Ppvernac.pr_vernac ast)
+    Richpp.richpp_of_pp ~width:78 (Ppvernac.pr_vernac ast)
 
 (** Goal display *)
 
@@ -169,60 +173,86 @@ let concl_next_tac =
     "right"
   ])
 
-let process_goal sigma g =
-  let env = Goal.V82.env sigma g in
+let process_goal short sigma g =
+  let evi = Evd.find sigma g in
+  let env = Evd.evar_filtered_env (Global.env ()) evi in
   let min_env = Environ.reset_context env in
-  let name = if Printer.print_goal_names () then Some (Names.Id.to_string (Termops.evar_suggested_name g sigma)) else None in
+  let name = if Printer.print_goal_names () then Some (Names.Id.to_string (Termops.evar_suggested_name env sigma g)) else None in
   let ccl =
-    pr_letype_env ~goal_concl_style:true env sigma (Goal.V82.concl sigma g)
+    pr_letype_env ~goal_concl_style:true env sigma (Evd.evar_concl evi)
   in
   let process_hyp d (env,l) =
     let d' = CompactedDecl.to_named_context d in
       (List.fold_right Environ.push_named d' env,
        (pr_compacted_decl env sigma d) :: l) in
-  let (_env, hyps) =
-    Context.Compacted.fold process_hyp
-      (Termops.compact_named_context (Environ.named_context env)) ~init:(min_env,[]) in
-  { Interface.goal_hyp = List.rev hyps; Interface.goal_ccl = ccl; Interface.goal_id = Goal.uid g; Interface.goal_name = name }
+  let hyps =
+    if short then [] else
+      let (_env, hyps) =
+        Context.Compacted.fold process_hyp
+        (Termops.compact_named_context (Environ.named_context env)) ~init:(min_env,[])
+      in
+      hyps
+  in
+  { Interface.goal_hyp = List.rev hyps; Interface.goal_ccl = ccl; Interface.goal_id = Proof.goal_uid g; Interface.goal_name = name }
 
 let process_goal_diffs diff_goal_map oldp nsigma ng =
-  let open Evd in
-  let name = if Printer.print_goal_names () then Some (Names.Id.to_string (Termops.evar_suggested_name ng nsigma)) else None in
-  let og_s = match oldp with
-    | Some oldp ->
-      let Proof.{ sigma=osigma } = Proof.data oldp in
-      (try Some { it = Evar.Map.find ng diff_goal_map; sigma = osigma }
-       with Not_found -> None)
-    | None -> None
+  let env = Global.env () in
+  let name = if Printer.print_goal_names () then Some (Names.Id.to_string (Termops.evar_suggested_name env nsigma ng)) else None in
+  let og_s = match oldp, diff_goal_map with
+  | Some oldp, Some diff_goal_map -> Proof_diffs.map_goal ng diff_goal_map
+  | None, _ | _, None -> None
   in
-  let (hyps_pp_list, concl_pp) = Proof_diffs.diff_goal_ide og_s ng nsigma in
-  { Interface.goal_hyp = hyps_pp_list; Interface.goal_ccl = concl_pp; Interface.goal_id = Goal.uid ng; Interface.goal_name = name }
+  let (hyps_pp_list, concl_pp) = Proof_diffs.diff_goal ?og_s (Proof_diffs.make_goal env nsigma ng) in
+  { Interface.goal_hyp = hyps_pp_list; Interface.goal_ccl = concl_pp;
+    Interface.goal_id = Proof.goal_uid ng; Interface.goal_name = name }
 
-let export_pre_goals Proof.{ sigma; goals; stack } process =
-  let process = List.map (process sigma) in
-  { Interface.fg_goals       = process goals
-  ; Interface.bg_goals       = List.(map (fun (lg,rg) -> process lg, process rg)) stack
-  ; Interface.shelved_goals  = process @@ Evd.shelf sigma
-  ; Interface.given_up_goals = process (Evar.Set.elements @@ Evd.given_up sigma)
-  }
+let export_pre_goals flags Proof.{ sigma; goals; stack } process =
+  let open Interface in
+  let process x = List.map (process sigma) x in
+  let fg_goals = if flags.gf_fg then process goals else [] in
+  let bg_goals =
+    if flags.gf_bg then List.(map (fun (lg,rg) -> process lg, process rg)) stack
+    else []
+  in
+  let shelved_goals =
+    if flags.gf_shelved then process @@ Evd.shelf sigma
+    else []
+  in
+  let given_up_goals =
+    if flags.gf_given_up then process (Evar.Set.elements @@ Evd.given_up sigma)
+    else []
+  in
+  { fg_goals; bg_goals; shelved_goals; given_up_goals }
 
-let goals () =
+let subgoals flags =
   let doc = get_doc () in
   set_doc @@ Stm.finish ~doc;
+  let short = match flags.Interface.gf_mode with
+  | "short" -> true
+  | _ -> false
+  in
   try
     let newp = Vernacstate.Declare.give_me_the_proof () in
     if Proof_diffs.show_diffs () then begin
       let oldp = Stm.get_prev_proof ~doc (Stm.get_current_state ~doc) in
       (try
-        let diff_goal_map = Proof_diffs.make_goal_map oldp newp in
-        Some (export_pre_goals Proof.(data newp) (process_goal_diffs diff_goal_map oldp))
+        let diff_goal_map = match oldp with
+        | None -> None
+        | Some oldp -> Some (Proof_diffs.make_goal_map oldp newp)
+        in
+        Some (export_pre_goals flags Proof.(data newp) (process_goal_diffs diff_goal_map oldp))
        with Pp_diff.Diff_Failure msg ->
          Proof_diffs.notify_proof_diff_failure msg;
-         Some (export_pre_goals Proof.(data newp) process_goal))
+         Some (export_pre_goals flags Proof.(data newp) (process_goal short)))
     end else
-      Some (export_pre_goals Proof.(data newp) process_goal)
+      Some (export_pre_goals flags Proof.(data newp) (process_goal short))
   with Vernacstate.Declare.NoCurrentProof -> None
-  [@@ocaml.warning "-3"];;
+  [@@ocaml.warning "-3"]
+
+let goals () =
+  let open Interface in
+  let all = { gf_mode = "full"; gf_fg = true; gf_bg = true; gf_shelved = true; gf_given_up = true } in
+  subgoals all
 
 let evars () =
   try
@@ -244,7 +274,7 @@ let hints () =
     match goals with
     | [] -> None
     | g :: _ ->
-      let env = Goal.V82.env sigma g in
+      let env = Evd.evar_filtered_env (Global.env ()) (Evd.find sigma g) in
       let get_hint_hyp env d accu = hyp_next_tac sigma env d :: accu in
       let hint_hyps = List.rev (Environ.fold_named_context get_hint_hyp env ~init: []) in
       Some (hint_hyps, concl_next_tac)
@@ -309,7 +339,7 @@ let dirpath_of_string_list s =
   let id =
     try Nametab.full_name_module qid
     with Not_found ->
-      CErrors.user_err ~hdr:"Search.interface_search"
+      CErrors.user_err
         (str "Module " ++ str path ++ str " not found.")
   in
   id
@@ -360,10 +390,33 @@ let proof_diff (diff_opt, sid) =
       let old = Stm.get_prev_proof ~doc sid in
       Proof_diffs.diff_proofs ~diff_opt ?old proof
 
-let debug_cmd = ref ""
+let debug_cmd = ref DebugHook.Action.Ignore
 
 let db_cmd cmd =
-  debug_cmd := cmd
+  debug_cmd := DebugHook.Action.Command cmd
+
+let db_continue opt =
+  let open DebugHook.Action in
+  debug_cmd := match opt with
+  | Interface.StepIn -> StepIn
+  | Interface.StepOver -> StepOver
+  | Interface.StepOut -> StepOut
+  | Interface.Continue -> Continue
+  | Interface.Interrupt -> Interrupt
+
+let db_upd_bpts updates =
+  debug_cmd := DebugHook.Action.UpdBpts updates
+
+let db_stack () =
+  debug_cmd := DebugHook.Action.GetStack;
+  []  (* return value passed through DebugHook.Answer *)
+
+let db_vars framenum =
+  debug_cmd := DebugHook.Action.GetVars framenum;
+  []  (* return value passed through DebugHook.Answer *)
+
+let db_configd () =
+  debug_cmd := DebugHook.Action.Configd
 
 let get_options () =
   let table = Goptions.get_tables () in
@@ -454,7 +507,12 @@ let eval_call c =
     Interface.status = interruptible status;
     Interface.search = interruptible search;
     Interface.proof_diff = interruptible proof_diff;
-    Interface.db_cmd = interruptible db_cmd;   (* todo: interruptible or not? *)
+    Interface.db_cmd = db_cmd;
+    Interface.db_upd_bpts = db_upd_bpts;
+    Interface.db_continue = db_continue;
+    Interface.db_stack = db_stack;
+    Interface.db_vars = db_vars;
+    Interface.db_configd = db_configd;
     Interface.get_options = interruptible get_options;
     Interface.set_options = interruptible set_options;
     Interface.mkcases = interruptible idetop_make_cases;
@@ -467,6 +525,7 @@ let eval_call c =
     Interface.stop_worker = Stm.stop_worker;
     Interface.print_ast = print_ast;
     Interface.annotate = interruptible annotate;
+    Interface.subgoals = interruptible subgoals;
   } in
   Xmlprotocol.abstract_eval_call handler c
 
@@ -495,8 +554,8 @@ let slave_feeder fmt xml_oc msg =
     trying to answer malformed requests. *)
 
 let msg_format = ref (fun () ->
-    let margin = Option.default 72 (Topfmt.get_margin ()) in
-    Xmlprotocol.Richpp margin
+    let width = Option.default 72 (Topfmt.get_margin ()) in
+    Xmlprotocol.Richpp { width; depth = max_int }
   )
 
 (* The loop ignores the command line arguments as the current model delegates
@@ -526,11 +585,14 @@ let loop ( { Coqtop.run_mode; color_mode },_) ~opts:_ state =
         pr_with_pid (Xml_printer.to_string_fmt xml_query);
       let Xmlprotocol.Unknown q = Xmlprotocol.to_call xml_query in
       let () = pr_debug_call q in
-      let r  = eval_call q in
-      let () = pr_debug_answer q r in
-(*       pr_with_pid (Xml_printer.to_string_fmt (Xmlprotocol.of_answer q r)); *)
-      print_xml xml_oc Xmlprotocol.(of_answer (!msg_format ()) q r);
-      flush out_ch
+      let (send, r)  = eval_call q in
+      (* conditional send so that db_stack and db_vars can reply through DebugHook.Answer *)
+      if send then begin
+        let () = pr_debug_answer q r in
+  (*       pr_with_pid (Xml_printer.to_string_fmt (Xmlprotocol.of_answer q r)); *)
+        print_xml xml_oc Xmlprotocol.(of_answer (!msg_format ()) q r);
+        flush out_ch
+      end
     with
       | Xml_parser.Error (Xml_parser.Empty, _) ->
         pr_debug "End of input, exiting gracefully.";
@@ -550,30 +612,38 @@ let loop ( { Coqtop.run_mode; color_mode },_) ~opts:_ state =
      xmlprotocol file doesn't depend on vernac thus it has no access
      to DebugHook data, we may want to fix that. *)
   let ltac_debug_answer ans =
-    let tag, msg =
-      let open DebugHook.Answer in
-      match ans with
-      | Prompt msg -> "prompt", msg
-      | Goal msg -> "goal", (str "Goal:" ++ fnl () ++ msg)
-      | Output msg -> "output", msg
+    let open DebugHook.Answer in
+    let open Xmlprotocol in
+    let xml = match ans with
+      | Prompt msg -> of_ltac_debug_answer ~tag:"prompt" msg
+      | Goal msg -> of_ltac_debug_answer ~tag:"goal" msg
+      | Output msg -> of_ltac_debug_answer ~tag:"output" msg
+      | Init -> of_ltac_debug_answer ~tag:"init" (str "")
+      | Vars vars -> of_vars vars;
+      | Stack s -> of_stack s
     in
-    print_xml xml_oc Xmlprotocol.(of_ltac_debug_answer ~tag msg) in
+    print_xml xml_oc xml in
 
   (* XXX: no need to have a ref here *)
   let ltac_debug_parse () =
     let raw_cmd =
+      debug_cmd := DebugHook.Action.Ignore;
       process_xml_msg xml_ic xml_oc out_ch;
       !debug_cmd
     in
     let open DebugHook in
-    match Action.parse raw_cmd with
-    | Ok act -> act
-    | Error error -> ltac_debug_answer (Answer.Output (str error)); Action.Failed
+    match raw_cmd with
+    | Action.Command cmd ->
+      (match Action.parse cmd with
+      | Ok act -> act
+      | Error error -> ltac_debug_answer (Answer.Output (str error)); Action.Failed)
+    | act -> act
   in
 
   DebugHook.Intf.(set
       { read_cmd = ltac_debug_parse
       ; submit_answer = ltac_debug_answer
+      ; isTerminal = false;
       });
 
   while not !quit do
@@ -594,7 +664,7 @@ let rec parse = function
        x :: parse rest
   | [] -> []
 
-let coqidetop_specific_usage = Usage.{
+let coqidetop_specific_usage = Boot.Usage.{
   executable_name = "coqidetop";
   extra_args = "";
   extra_options = "\n\
