@@ -19,8 +19,8 @@ open CmdArgs.Prefs
 
 let (/) = Filename.concat
 
-let coq_version = "8.15.1"
-let vo_magic = 81500
+let coq_version = "8.16.0"
+let vo_magic = 81600
 let is_a_released_version = true
 
 (** Default OCaml binaries *)
@@ -42,10 +42,12 @@ let coq_strict_sequence = "-strict-sequence"
 let arch prefs = arch prefs.arch
 
 (** NB: [arch_is_win32] is broader than [os_type_win32], cf. cygwin *)
-
 let arch_is_win32 arch = arch = "win32"
 
-let resolve_binary_suffix arch = if arch_is_win32 arch then ".exe" else ""
+let unix arch = os_type_cygwin || not (arch_is_win32 arch)
+
+(* For cygwin dune doesn't add .exe to the binaries, only under mingw *)
+let resolve_binary_suffix arch = if unix arch then "" else ".exe"
 
 (** * Git Precommit Hook *)
 let install_precommit_hook prefs =
@@ -104,19 +106,19 @@ let caml_version_nums { CamlConf.caml_version } =
   generic_version_nums ~name:"the OCaml compiler" caml_version
 
 let check_caml_version prefs caml_version caml_version_nums =
-  if caml_version_nums >= [4;5;0] then
+  if caml_version_nums >= [4;9;0] then
     cprintf prefs "You have OCaml %s. Good!" caml_version
   else
     let () = cprintf prefs "Your version of OCaml is %s." caml_version in
-    die "You need OCaml 4.05.0 or later."
+    die "You need OCaml 4.09.0 or later."
 
 let check_findlib_version prefs { CamlConf.findlib_version } =
   let findlib_version_nums = generic_version_nums ~name:"findlib" findlib_version in
-  if findlib_version_nums >= [1;8;0] then
+  if findlib_version_nums >= [1;8;1] then
     cprintf prefs "You have OCamlfind %s. Good!" findlib_version
   else
     let () = cprintf prefs "Your version of OCamlfind is %s." findlib_version in
-    die "You need OCamlfind 1.8.0 or later."
+    die "You need OCamlfind 1.8.1 or later."
 
 (** Note, these warnings are only used in Coq Makefile *)
 (** Explanation of enabled/disabled warnings:
@@ -216,10 +218,10 @@ let check_for_zarith prefs =
     die "Zarith library installed but no development files found (try installing the -dev package)"
   | _   ->
     let zarith_version_int = generic_version_nums ~name:"Zarith" zarith_version in
-    if zarith_version_int >= [1;10;0] then
+    if zarith_version_int >= [1;11;0] then
       cprintf prefs "You have the Zarith library %s installed. Good!" zarith_version
     else
-      die ("Zarith version 1.10 is required, you have " ^ zarith_version)
+      die ("Zarith version 1.11 is required, you have " ^ zarith_version)
 
 (** * Documentation : do we have latex, hevea, ... *)
 
@@ -239,8 +241,6 @@ let check_doc () =
 (* Source code root *)
 let coqsrc = Sys.getcwd ()
 
-let unix arch = os_type_cygwin || not (arch_is_win32 arch)
-
 (** Variable name, description, ref in prefs, default dir, prefix-relative *)
 
 type path_style =
